@@ -4,7 +4,8 @@ import numpy as np
 from napari_plugin_engine import napari_hook_implementation
 from napari_tools_menu import register_function
 import napari
-
+from napari.types import LayerDataTuple
+from typing import List
 
 # This is the actual plugin function, where we export our function
 # (The functions themselves are defined below)
@@ -13,12 +14,12 @@ def napari_experimental_provide_function():
     return [crop_region]
 
 
-@register_function(menu="Utilities > Crop region")
+@register_function(menu="Utilities > Crop region(s)")
 def crop_region(
     layer: napari.layers.Layer,
     shapes_layer: napari.layers.Shapes,
-) -> napari.layers.Layer:
-
+) -> List[LayerDataTuple]:
+    """Crop regions in napari defined by shapes."""
     if shapes_layer is None:
         shapes_layer.mode = "add_rectangle"
         warnings.warn("Please annotate a region to crop.")
@@ -33,50 +34,73 @@ def crop_region(
 
     layer_data, layer_props, layer_type = layer.as_layer_data_tuple()
 
+    try:
+        rgb = layer_props["rgb"]
+    except KeyError:
+        rgb = False
+
     shape_types = shapes_layer.shape_type
     shapes = shapes_layer.data
-    for shape, shape_type in zip(shapes, shape_types):
-        # removes an artifact when creating an ellipse mask
-        shape = np.ceil(shape)
+    cropped_list = []
+    for count, [shape, shape_type] in enumerate(zip(shapes, shape_types)):
         # move shape vertices to within image coordinate limits
         layer_shape = np.array(layer_data.shape)
-        if layer_props["rgb"]:
+        if rgb:
             layer_shape = layer_shape[:-1]
-        shape = np.max([shape, np.zeros(shape.shape)], axis=0)
-        shape = np.min([shape, np.resize(layer_shape, shape.shape)], axis=0)
         # find min and max for each dimension
-        start = np.min(shape, axis=0)
-        stop = np.max(shape, axis=0)
+        start = np.rint(np.min(shape, axis=0))
+        stop = np.rint(np.max(shape, axis=0))
         # create slicing indices
         slices = tuple(
             slice(first, last + 1) if first != last else slice(0, None)
-            for first, last in np.stack([start, stop]).astype(int).T
+            for first, last in np.stack([start.clip(0),
+                                         stop.clip(0)]).astype(int).T
         )
         cropped_data = layer_data[slices].copy()
         # handle polygons
         if shape_type != "rectangle":
             mask_nD_shape = np.array(
-                [1 if slc.stop == None else (slc.stop - slc.start) for slc in slices]
+                [1 if slc.stop is None
+                 else (min(slc.stop, layer_data.shape[i]) - slc.start)
+                 for i, slc in enumerate(slices)]
             )
-            # remove extra dimensions from shape vertices (draw in a single plane)
-            verts_flat = np.array(shape - start)[:, mask_nD_shape > 1]
+            # remove extra dimensions from shape vertices
+            # (draw in a single plane)
+            verts_flat = np.array(shape - start.clip(0))[:, mask_nD_shape > 1]
             # get a 2D mask
             mask_2D = (
                 napari.layers.Shapes(verts_flat, shape_type=shape_type)
                 .to_masks()
                 .squeeze()
             )
+            # match cropped_data (x,y) shape with mask_2D shape
+            if rgb:
+                shape_dif_2D = np.array(cropped_data.shape[-3:-1]) \
+                    - np.array(mask_2D.shape)
+            else:
+                shape_dif_2D = np.array(cropped_data.shape[-2:]) \
+                    - np.array(mask_2D.shape)
+            shape_dif_2D = [None if i == 0 else i
+                            for i in shape_dif_2D.tolist()]
+            mask_2D = mask_2D[:shape_dif_2D[-2], :shape_dif_2D[-1]]
             # add back the rgb(a) dimension
-            if layer_props["rgb"]:
+            if rgb:
                 mask_nD_shape = np.append(mask_nD_shape, 1)
             # add back dimensions of the original vertices
             mask_nD = mask_2D.reshape(mask_nD_shape)
             # broadcast the mask to the shape of the cropped image
             mask = np.broadcast_to(mask_nD, cropped_data.shape)
+            # erase pixels outside drawn shape
             cropped_data[~mask] = 0
 
-    layer_props["name"] = layer_props["name"] + " (cropped)"
-    if layer_type == "image":
-        return napari.layers.Image(cropped_data, **layer_props)
-    if layer_type == "labels":
-        return napari.layers.Labels(cropped_data, **layer_props)
+            # trim zeros
+            non_zero = np.where(cropped_data != 0)
+            indices = [slice(min(i_nz), max(i_nz) + 1) for i_nz in non_zero]
+            if rgb:
+                indices[-1] = slice(None)
+            cropped_data = cropped_data[tuple(indices)]
+
+        new_layer_props = layer_props.copy()
+        new_layer_props["name"] = layer_props["name"] + f" (cropped {count+1})"
+        cropped_list.append((cropped_data, new_layer_props, layer_type))
+    return cropped_list
